#include <linux/proc_fs.h>

#include "request.h"

static ssize_t secumod_input(struct file *file, const char *buf, size_t length, loff_t *offset)
  /* TODO: do better logging, fix pointer errors after deleting paths */
  /* ( if you want kfree to become lazy, sort the calls to kmalloc and kfree ) */
{
	int len, i, j;
	char input[sizeof(PATH_ENTRY)]; 
	char ** sel; // pointer to array of pointers to strings
	int * cnt;   // pointer to counter
	PATH_ENTRY entry;

	for(len = 0; len < sizeof(PATH_ENTRY); len++) {
		get_user(input[len], buf+len);
	}

	memcpy(&entry, input, sizeof(PATH_ENTRY));
	
	sel = texec_paths;
	cnt = &path_count;
	if (entry.request_type==SUID_REQUEST) 
	{
		sel = suid_paths;
		cnt = &suid_count;
	};
	if (entry.request_type==TEXEC_REQUEST)
	{
		sel = texec_paths;
		cnt = &path_count;
	};

	printk(NOTIFY_LOG "SecuMod path interface: request group %d type %d\n",
	       entry.request_type, entry.type);

	switch (entry.type) {
		case PATH_INSERT:
			if (path_count < NR_TPATHS) {
				j=-1; 
				printk(NOTIFY_LOG "SecuMod about to add a trusted path");

//			       	printk(NOTIFY_LOG "SecuMod shall add path %.240s (%d) to type %d trust\n",
//			       	       entry.path,strlen(entry.path),entry.request_type);

			        for (i=0; (i<NR_TPATHS) && (!j); i++)
			        	if (!sel[i]) j=i; /* look for a free slot */
			        if (j>=0)  /* should always be the case */
			        {
				if (( sel[j] = 
				      (char *)kmalloc(strlen(entry.path)+2, GFP_KERNEL) ))
				  {  
				  strncpy(sel[j], entry.path, strlen(entry.path)+1);
				  (*cnt)++; /* strlen zaehlt die \0 nicht, strncpy schon. */
				  printk(NOTIFY_LOG "SecuMod path slot %d @ %p now %.240s\n",
				         j,&(sel[j][0]),sel[j]);
				  };
				}; /* TODO: stop using kmalloc */
			}
			break;		

		case PATH_REMOVE:
			printk(NOTIFY_LOG "SecuMod about to remove a trusted path");
			for (i = 0; i < NR_TPATHS; i++) if (sel[i]) {
				if (strncmp(entry.path, sel[i], strlen(sel[i])) == 0) {
					kfree((void *)sel[i]);
					sel[i] = NULL;
					(*cnt)--;
				}
			}
			break;

		case PATH_CLEAR: // should be called before module is unloaded
			for (i = 0; i < NR_TPATHS; i++) {
				if (sel[i]) kfree((void *)sel[i]);
				sel[i] = NULL;
			}
			(*cnt) = 0;
			printk(NOTIFY_LOG "SecuMod will only trust "
			       "built-in paths for type %d now\n", 
			       entry.request_type);
			break;

		case PATH_SHOW:
		        printk(REAL_LOG "SecuMod trusted paths (type %d) are: [\n", 
		               entry.request_type);
			for (i = 0; i < NR_TPATHS; i++) if (sel[i]) {
				printk(REAL_LOG "%d @ %p: %.240s\n", 
				       i, &(sel[i][0]), sel[i]);
				/* strange way to do the output */
			}
			printk(REAL_LOG "] SecuMod supports built-in paths too, be aware\n");
			break;
	}

	return len;
}

static int secumod_permission(struct inode *inode, int op)
{
	if (op == 4 || (op == 2 && current->euid == 0))
		return 0; 

	return -EACCES;
}


int secumod_open(struct inode *inode, struct file *file)
{
	return 0;
}

int secumod_close(struct inode *inode, struct file *file)
{
	return 0;
}

static struct file_operations secumod_fileop = {
	NULL,
	NULL,
	secumod_input,
	NULL,
	NULL,
	NULL,
	NULL,
	secumod_open,
	NULL,
	secumod_close,
};

static struct inode_operations secumod_inodop = {
	&secumod_fileop,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	secumod_permission
};

static struct proc_dir_entry secumod_procfile = {
	0,
	7,
	"secumod",
	S_IFREG | S_IRUGO | S_IWUSR, 
	1,
	0, 0,
	0,
	&secumod_inodop, 
	NULL  
}; 
