unsigned char syscall_bytes[256][16]; 
/* TODO: do not waste memory for multiple sys_ni_syscall instances */
/* TODO: find why only access to c0000000..c0000000+ramsize works  */
/* probably the crash only occurs for syscalls pointing INSIDE secumod? */
/* (INSIDE secumod == c0000000 + ramsize + a few kB during init_module) */

volatile int table_stored = 0;

/* backup stuff */
void make_sys_call_table_backup()
{
  int counter,counter2;
  unsigned char * ptr;

  printk(NOTIFY_LOG "Secumod will now backup the sys_call_table\n");
  if (table_stored) printk(NOTIFY_LOG "Secumod will not do backup more than once!\n");

  if (!table_stored)
  for (counter = 0; counter < 255; counter++) 
  {
    ptr = sys_call_table_backup[counter] = sys_call_table[counter];
    if ((ptr < (unsigned char *)0xc0001000L) /* TODO: use (>1GB) compile opts */
     || (ptr > (unsigned char *)0xc1000000L)) /* be careful if 'Kernel' > 16 MB */
    {  printk(REAL_LOG "Secumod: will not backup bytes of syscall %u @ %p\n",
              counter,ptr);
    } else
    {  begin_kmem();
       for (counter2=0; counter2<16; counter2++)
         syscall_bytes[counter][counter2]=ptr[counter2];
       end_kmem();
#ifdef DUMP_SCT       
       printk(NOTIFY_LOG "Syscall %u @ %p starts with %.2x %.2x %.2x %.2x  %.2x %.2x %.2x %.2x\n",
        counter, ptr,
        syscall_bytes[counter][0], syscall_bytes[counter][1], syscall_bytes[counter][2],
        syscall_bytes[counter][3], syscall_bytes[counter][4], syscall_bytes[counter][5],
        syscall_bytes[counter][6], syscall_bytes[counter][7]);
#endif
    };
  };

  table_stored++;
  printk(NOTIFY_LOG "Secumod has made a backup of the sys_call_table\n");
  return;
}

/* the checking routine */
void check_sys_call_table_backup(int verbose)
{
  int counter,counter2;
  void * ptr;

  if (verbose) printk(NOTIFY_LOG "Secumod will now check the sys_call_table\n");

  if (table_stored)
  for (counter = 0; counter < 255; counter++) 
  {
    ptr = sys_call_table[counter];
    if (ptr != sys_call_table_backup[counter]) 
    {
      if (logging)
        printk(REAL_LOG "SecuMod: Syscall Entry %u was modified\n",
               counter);
      /* restore */
      // sys_call_hooked is not checked, so take care to do the backup AFTER the hook!
      sys_call_table[counter] = sys_call_table_backup[counter];
      if (logging)
        printk(REAL_LOG "SecuMod: replace %u with backuped system call\n", 
               counter);
      
    } else /* entry unmodified */
    { /* abtrom check */
      if  ((ptr < (void *)0xc0001000L) /* TODO: use (>1GB) compile opts */
        || (ptr > (void *)0xc1000000L)) /* be careful if 'Kernel' > 16 MB */
      {  if (verbose) printk(NOTIFY_LOG "Secumod: will not check bytes of syscall %u @ %p\n",
                             counter,ptr);
      } else
      {  int changed;
         changed = 0;
         begin_kmem();
         for (counter2=0; counter2<16; counter2++)
           if (syscall_bytes[counter][counter2]!=
               ((unsigned char * )ptr)[counter2]) changed++;
         end_kmem();
         if (changed)
         {
           if (logging)
             printk(REAL_LOG "Secumod: at least %u bytes changed in system call %u @ %p\n",
                    changed, counter, ptr);
           /* TODO: try to remove abtrom type hooks? will be VERY hard! */
         };
      };
    }; /* end of abtrom check */
  };
  
  if (verbose)
  { if (table_stored)
    { printk(NOTIFY_LOG "Secumod has checked the sys_call_table\n");
    } else 
    { printk(NOTIFY_LOG "Secumod could not check s.c.t. - no reference stored yet\n");
    };
  };
  return;
};


int (*orig_init_module)(const char *name_user, struct module *mod_user);
                         

int hacked_init_module(const char *name_user, struct module *mod_user) 
{
	int ret;

	/* execute module_init() */
	ret = orig_init_module(name_user, mod_user);

	check_sys_call_table_backup(1); // 1 = verbose

	return ret;
};


void init_module_protection(int on)
{
	init_module_protection_on = on;
	if (on) 
	{
		orig_init_module = sys_call_table[SYS_init_module];
		sys_call_table[SYS_init_module] = hacked_init_module;
		sys_call_hooked[SYS_init_module] = 1;
	} else 
	{ 
		if (sys_call_hooked[SYS_init_module])
			sys_call_table[SYS_init_module] = orig_init_module;
		sys_call_hooked[SYS_init_module] = 0; 
	};
};
